import React from 'react';

var classCallCheck = function (instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
};

var createClass = function () {
  function defineProperties(target, props) {
    for (var i = 0; i < props.length; i++) {
      var descriptor = props[i];
      descriptor.enumerable = descriptor.enumerable || false;
      descriptor.configurable = true;
      if ("value" in descriptor) descriptor.writable = true;
      Object.defineProperty(target, descriptor.key, descriptor);
    }
  }

  return function (Constructor, protoProps, staticProps) {
    if (protoProps) defineProperties(Constructor.prototype, protoProps);
    if (staticProps) defineProperties(Constructor, staticProps);
    return Constructor;
  };
}();

var defineProperty = function (obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
};

var _extends = Object.assign || function (target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i];

    for (var key in source) {
      if (Object.prototype.hasOwnProperty.call(source, key)) {
        target[key] = source[key];
      }
    }
  }

  return target;
};

var toConsumableArray = function (arr) {
  if (Array.isArray(arr)) {
    for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) arr2[i] = arr[i];

    return arr2;
  } else {
    return Array.from(arr);
  }
};

/**
 * BaseArrayReducerMethods
 * To overrides these methods just create a class that extends to this class
 */
var BaseArrayReducerMethods = function () {
  function BaseArrayReducerMethods() {
    classCallCheck(this, BaseArrayReducerMethods);

    this.receiveMany = function (state, payload) {
      return [].concat(toConsumableArray(state), toConsumableArray(payload));
    };

    this.receiveManyOnTop = function (state, payload) {
      return [].concat(toConsumableArray(payload), toConsumableArray(state));
    };

    this.substituteOne = function (state, payload) {
      var indexOfPayload = state.findIndex(function (item) {
        return item.id === payload.id;
      });
      if (indexOfPayload > -1) {
        return BaseArrayReducerMethods._update(indexOfPayload, payload, state);
      }
      return state;
    };

    this.destroyOne = function (state, payload) {
      return state.filter(function (item) {
        return item.id !== payload;
      });
    };

    this.destroyMany = function (state, payload) {
      return state.filter(function (item) {
        return !payload.some(function (id) {
          return id === item.id;
        });
      });
    };

    this.resetAll = function (state, payload) {
      return [];
    };
  }

  createClass(BaseArrayReducerMethods, null, [{
    key: "_update",
    value: function _update(index, value, array) {
      return [].concat(toConsumableArray(array.slice(0, index)), [value], toConsumableArray(array.slice(index + 1)));
    }

    /**
     * receiveMany
     * @param {*[] | *} state
     * @param {*[] | *} payload
     * @return {*[] | *}
     */


    /**
     * substituteOne
     * @param {*[] | *} state
     * @param {object | *} payload
     * @returns {*[]}
     */


    /**
     * destroyMany
     * @param {*[] | *} state
     * @param {string[]} payload
     * @returns {*[] | *}
     */

  }]);
  return BaseArrayReducerMethods;
}();

var baseArrayReducer = function baseArrayReducer(section) {
  var initialState = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
  var Overrides = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : BaseArrayReducerMethods;
  var reducer = arguments[3];
  return function () {
    var state = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : initialState;
    var action = arguments[1];

    var methods = new Overrides();

    switch (action.type) {
      case "RECEIVE_" + section + "S":
        return methods.receiveMany(state, action.payload);
      case "RECEIVE_" + section + "S_ON_TOP":
        return methods.receiveManyOnTop(state, action.payload);
      case "SUBSTITUTE_" + section:
        return methods.substituteOne(state, action.payload);
      case "DESTROY_" + section:
        return methods.destroyOne(state, action.payload);
      case "DESTROY_" + section + "S":
        return methods.destroyMany(state, action.payload);
      case "RESET_" + section + "S":
        return methods.resetAll(state, action.payload);
      default:
        return reducer ? reducer(state, action) : state;
    }
  };
};

/**
 * createActions
 * @param {String} section
 * @param {Object} options you can inject additional types and action
 * Create a template with common actions and types. Just need to pass a section name
 * and register it into asyncReducersActions.js file
 */
var createActions = function createActions(section) {
  var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : { types: {}, actions: {} };
  return {
    types: _extends({
      FIND: "FIND_" + section + "S",
      FIND_BY_ID: "FIND_" + section + "_BY_ID",
      CREATE: "CREATE_" + section,
      UPDATE_BY_ID: "UPDATE_" + section + "_BY_ID",
      DELETE_BY_ID: "DELETE_" + section + "_BY_ID",
      RECEIVE_MANY: "RECEIVE_" + section + "S",
      RECEIVE_MANY_ON_TOP: "RECEIVE_" + section + "S_ON_TOP",
      RECEIVE_ONE: "RECEIVE_" + section,
      RECEIVE_ONE_ON_TOP: "RECEIVE_" + section + "_ON_TOP",
      SUBSTITUTE_ONE: "SUBSTITUTE_" + section,
      DESTROY_ONE: "DESTROY_" + section,
      DESTROY_MANY: "DESTROY_" + section + "S",
      RESET: "RESET_" + section + "S"
    }, options.types),

    // The only way to reference the same Object's properties during its creation is to use a getter
    get actions() {
      var _this = this,
          _babelHelpers$extends;

      return _extends((_babelHelpers$extends = {}, defineProperty(_babelHelpers$extends, "find" + this._normalizeName() + "s", function undefined(urlObject, callback) {
        return {
          type: _this.types.FIND,
          payload: { urlObject: urlObject, callback: callback }
        };
      }), defineProperty(_babelHelpers$extends, "find" + this._normalizeName() + "ById", function undefined(urlObject, callback) {
        return {
          type: _this.types.FIND_BY_ID,
          payload: { urlObject: urlObject, callback: callback }
        };
      }), defineProperty(_babelHelpers$extends, "create" + this._normalizeName(), function undefined(urlObject, data, callback) {
        return {
          type: _this.types.CREATE,
          payload: { urlObject: urlObject, data: data, callback: callback }
        };
      }), defineProperty(_babelHelpers$extends, "update" + this._normalizeName() + "ById", function undefined(urlObject, data, callback) {
        return {
          type: _this.types.UPDATE_BY_ID,
          payload: { urlObject: urlObject, data: data, callback: callback }
        };
      }), defineProperty(_babelHelpers$extends, "delete" + this._normalizeName() + "ById", function undefined(urlObject, data, callback) {
        return {
          type: _this.types.DELETE_BY_ID,
          payload: { urlObject: urlObject, data: data, callback: callback }
        };
      }), defineProperty(_babelHelpers$extends, "receive" + this._normalizeName() + "s", function undefined(data, name) {
        return {
          type: _this.types.RECEIVE_MANY,
          name: name,
          payload: data
        };
      }), defineProperty(_babelHelpers$extends, "receive" + this._normalizeName() + "sOnTop", function undefined(data, name) {
        return {
          type: _this.types.RECEIVE_MANY_ON_TOP,
          name: name,
          payload: data
        };
      }), defineProperty(_babelHelpers$extends, "receive" + this._normalizeName(), function undefined(data, name) {
        return {
          type: _this.types.RECEIVE_ONE,
          name: name,
          payload: data
        };
      }), defineProperty(_babelHelpers$extends, "receive" + this._normalizeName() + "OnTop", function undefined(data, name) {
        return {
          type: _this.types.RECEIVE_ONE_ON_TOP,
          name: name,
          payload: data
        };
      }), defineProperty(_babelHelpers$extends, "substitute" + this._normalizeName(), function undefined(data, name) {
        return {
          type: _this.types.SUBSTITUTE_ONE,
          name: name,
          payload: data
        };
      }), defineProperty(_babelHelpers$extends, "destroy" + this._normalizeName(), function undefined(data, name) {
        return {
          type: _this.types.DESTROY_MANY,
          name: name,
          payload: data
        };
      }), defineProperty(_babelHelpers$extends, "destroy" + this._normalizeName() + "s", function undefined(data, name) {
        return {
          type: _this.types.DESTROY_MANY,
          name: name,
          payload: data
        };
      }), defineProperty(_babelHelpers$extends, "reset" + this._normalizeName() + "s", function undefined(name) {
        return {
          type: _this.types.RESET,
          name: name
        };
      }), _babelHelpers$extends), options.actions);
    },

    /**
     * _normalizeName
     * @returns {string}
     * @private
     * input "EVENT_AVAILABLE" -> output "EventAvailable"
     */
    _normalizeName: function _normalizeName() {
      var capitalizeFirstLetter = function capitalizeFirstLetter(name) {
        var firstCapitalLetter = name[0];
        var restOfTheNameLowerCase = name.slice(1).toLowerCase();

        return firstCapitalLetter + restOfTheNameLowerCase;
      };

      if (section.includes("_")) {
        return section.split("_").map(function (name) {
          return capitalizeFirstLetter(name);
        }).join("");
      } else {
        return capitalizeFirstLetter(section);
      }
    }
  };
};

var BaseObjectReducerMethods = function BaseObjectReducerMethods() {
  classCallCheck(this, BaseObjectReducerMethods);

  this.receiveOne = function (state, payload) {
    return payload;
  };
};

var baseObjectReducer = function baseObjectReducer(section) {
  var Overrides = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : BaseObjectReducerMethods;
  var reducer = arguments[3];
  return function () {
    var state = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    var action = arguments[1];

    var methods = new Overrides();

    if (action.type === "RECEIVE_" + section) {
      return methods.receiveOne(state, action.payload);
    } else {
      return reducer ? reducer(state, action) : state;
    }
  };
};

/**
 * _searchProperty
 * it search through props object or to params object
 * @param {object} props
 * @param {string} name
 * @returns {*}
 */
function searchProperty(props, name) {
  return props[name] || props.match && props.match.params[name];
}

/**
 * _getNamesFromIdentifierName
 * This method convert identifierName in array
 * @param {string | string[]} identifierName
 * @return {array}
 */
var getArrayFromIdentifierNames = function getArrayFromIdentifierNames(identifierName) {
  return Array.isArray(identifierName) ? identifierName : [identifierName];
};

/**
 * mapIdentifierNamesToProp
 * @param {string[]} identifierNames
 * @param {object} props
 * @return {string[]}
 */
function mapIdentifierNamesToProp(identifierNames, props) {
  return identifierNames.map(function (name) {
    return searchProperty(props, name);
  });
}

/**
 * getReducerName
 * @param {object} props
 * @param {string[] | string} identifierName
 * @param {string} namespace
 * @return {string}
 */
function getReducerName(props, identifierName, namespace) {
  if (identifierName) {
    var identifiersArray = getArrayFromIdentifierNames(identifierName);
    var mappedProps = mapIdentifierNamesToProp(identifiersArray, props);

    return namespace + "/" + mappedProps.join("/");
  }

  return namespace;
}

// This function check the reducer name from the action and instantiate a reducer
function createNamedWrapperReducer(reducerFunction) {
  return function (reducerName) {
    return function (state, action) {
      var name = action.name;

      var isInitializationCall = state === undefined;
      if (name !== reducerName && !isInitializationCall) return state;

      return reducerFunction(state, action);
    };
  };
}

/**
 * createAnonymousDuck
 * @param {string} name
 * @param {string} section
 * @param {string} type
 */
function createDuck(name, section, type) {
  var reducerType = {
    array: baseArrayReducer,
    object: baseObjectReducer
  };
  var reducer = reducerType[type](section);

  return _extends({
    reducer: createNamedWrapperReducer(reducer)(name)
  }, createActions(section));
}

/* eslint-disable react/prop-types */

var ReactReduxOnTheFlyContext = React.createContext({});

function ReactReduxOnTheFlyProvider(_ref) {
  var children = _ref.children,
      store = _ref.store,
      reactReduxOnTheFly = _ref.reactReduxOnTheFly;

  return React.createElement(
    ReactReduxOnTheFlyContext.Provider,
    { value: { store: store, reactReduxOnTheFly: reactReduxOnTheFly } },
    children
  );
}

/* eslint-disable react/prop-types */

/**
 * withAsyncReducer
 * HOC to inject a reducer into the store dynamically
 * @param {string} namespace
 * @param {string | string[]} [identifierName] id or any other unique identifier
 * @param anonymousReducer a reducer without name, the name will be generated on component injection
 */
var withAsyncReducer = function withAsyncReducer(namespace, identifierName, anonymousReducer) {
  return function (WrappedComponent) {
    return function (props) {
      var reducerName = getReducerName(props, identifierName, namespace);

      var reducer = anonymousReducer(reducerName);

      return React.createElement(
        ReactReduxOnTheFlyContext.Consumer,
        null,
        function (_ref) {
          var store = _ref.store,
              reactReduxOnTheFly = _ref.reactReduxOnTheFly;

          reactReduxOnTheFly.injectAsyncReducer(store, reducerName, reducer);
          reactReduxOnTheFly.asyncReducerManager(store, props);

          return React.createElement(WrappedComponent, _extends({}, props, { reducerName: reducerName }));
        }
      );
    };
  };
};

function symbolObservablePonyfill(root) {
	var result;
	var Symbol = root.Symbol;

	if (typeof Symbol === 'function') {
		if (Symbol.observable) {
			result = Symbol.observable;
		} else {
			result = Symbol('observable');
			Symbol.observable = result;
		}
	} else {
		result = '@@observable';
	}

	return result;
}

/* global window */

var root;

if (typeof self !== 'undefined') {
  root = self;
} else if (typeof window !== 'undefined') {
  root = window;
} else if (typeof global !== 'undefined') {
  root = global;
} else if (typeof module !== 'undefined') {
  root = module;
} else {
  root = Function('return this')();
}

var result = symbolObservablePonyfill(root);

/**
 * These are private action types reserved by Redux.
 * For any unknown actions, you must return the current state.
 * If the current state is undefined, you must return the initial state.
 * Do not reference these action types directly in your code.
 */
var randomString = function randomString() {
  return Math.random().toString(36).substring(7).split('').join('.');
};

var ActionTypes = {
  INIT: "@@redux/INIT" + randomString(),
  REPLACE: "@@redux/REPLACE" + randomString(),
  PROBE_UNKNOWN_ACTION: function PROBE_UNKNOWN_ACTION() {
    return "@@redux/PROBE_UNKNOWN_ACTION" + randomString();
  }
};

/**
 * @param {any} obj The object to inspect.
 * @returns {boolean} True if the argument appears to be a plain object.
 */
function isPlainObject(obj) {
  if (typeof obj !== 'object' || obj === null) return false;
  var proto = obj;

  while (Object.getPrototypeOf(proto) !== null) {
    proto = Object.getPrototypeOf(proto);
  }

  return Object.getPrototypeOf(obj) === proto;
}

/**
 * Creates a Redux store that holds the state tree.
 * The only way to change the data in the store is to call `dispatch()` on it.
 *
 * There should only be a single store in your app. To specify how different
 * parts of the state tree respond to actions, you may combine several reducers
 * into a single reducer function by using `combineReducers`.
 *
 * @param {Function} reducer A function that returns the next state tree, given
 * the current state tree and the action to handle.
 *
 * @param {any} [preloadedState] The initial state. You may optionally specify it
 * to hydrate the state from the server in universal apps, or to restore a
 * previously serialized user session.
 * If you use `combineReducers` to produce the root reducer function, this must be
 * an object with the same shape as `combineReducers` keys.
 *
 * @param {Function} [enhancer] The store enhancer. You may optionally specify it
 * to enhance the store with third-party capabilities such as middleware,
 * time travel, persistence, etc. The only store enhancer that ships with Redux
 * is `applyMiddleware()`.
 *
 * @returns {Store} A Redux store that lets you read the state, dispatch actions
 * and subscribe to changes.
 */

function createStore(reducer, preloadedState, enhancer) {
  var _ref2;

  if (typeof preloadedState === 'function' && typeof enhancer === 'function' || typeof enhancer === 'function' && typeof arguments[3] === 'function') {
    throw new Error('It looks like you are passing several store enhancers to ' + 'createStore(). This is not supported. Instead, compose them ' + 'together to a single function');
  }

  if (typeof preloadedState === 'function' && typeof enhancer === 'undefined') {
    enhancer = preloadedState;
    preloadedState = undefined;
  }

  if (typeof enhancer !== 'undefined') {
    if (typeof enhancer !== 'function') {
      throw new Error('Expected the enhancer to be a function.');
    }

    return enhancer(createStore)(reducer, preloadedState);
  }

  if (typeof reducer !== 'function') {
    throw new Error('Expected the reducer to be a function.');
  }

  var currentReducer = reducer;
  var currentState = preloadedState;
  var currentListeners = [];
  var nextListeners = currentListeners;
  var isDispatching = false;

  function ensureCanMutateNextListeners() {
    if (nextListeners === currentListeners) {
      nextListeners = currentListeners.slice();
    }
  }
  /**
   * Reads the state tree managed by the store.
   *
   * @returns {any} The current state tree of your application.
   */


  function getState() {
    if (isDispatching) {
      throw new Error('You may not call store.getState() while the reducer is executing. ' + 'The reducer has already received the state as an argument. ' + 'Pass it down from the top reducer instead of reading it from the store.');
    }

    return currentState;
  }
  /**
   * Adds a change listener. It will be called any time an action is dispatched,
   * and some part of the state tree may potentially have changed. You may then
   * call `getState()` to read the current state tree inside the callback.
   *
   * You may call `dispatch()` from a change listener, with the following
   * caveats:
   *
   * 1. The subscriptions are snapshotted just before every `dispatch()` call.
   * If you subscribe or unsubscribe while the listeners are being invoked, this
   * will not have any effect on the `dispatch()` that is currently in progress.
   * However, the next `dispatch()` call, whether nested or not, will use a more
   * recent snapshot of the subscription list.
   *
   * 2. The listener should not expect to see all state changes, as the state
   * might have been updated multiple times during a nested `dispatch()` before
   * the listener is called. It is, however, guaranteed that all subscribers
   * registered before the `dispatch()` started will be called with the latest
   * state by the time it exits.
   *
   * @param {Function} listener A callback to be invoked on every dispatch.
   * @returns {Function} A function to remove this change listener.
   */


  function subscribe(listener) {
    if (typeof listener !== 'function') {
      throw new Error('Expected the listener to be a function.');
    }

    if (isDispatching) {
      throw new Error('You may not call store.subscribe() while the reducer is executing. ' + 'If you would like to be notified after the store has been updated, subscribe from a ' + 'component and invoke store.getState() in the callback to access the latest state. ' + 'See https://redux.js.org/api-reference/store#subscribe(listener) for more details.');
    }

    var isSubscribed = true;
    ensureCanMutateNextListeners();
    nextListeners.push(listener);
    return function unsubscribe() {
      if (!isSubscribed) {
        return;
      }

      if (isDispatching) {
        throw new Error('You may not unsubscribe from a store listener while the reducer is executing. ' + 'See https://redux.js.org/api-reference/store#subscribe(listener) for more details.');
      }

      isSubscribed = false;
      ensureCanMutateNextListeners();
      var index = nextListeners.indexOf(listener);
      nextListeners.splice(index, 1);
    };
  }
  /**
   * Dispatches an action. It is the only way to trigger a state change.
   *
   * The `reducer` function, used to create the store, will be called with the
   * current state tree and the given `action`. Its return value will
   * be considered the **next** state of the tree, and the change listeners
   * will be notified.
   *
   * The base implementation only supports plain object actions. If you want to
   * dispatch a Promise, an Observable, a thunk, or something else, you need to
   * wrap your store creating function into the corresponding middleware. For
   * example, see the documentation for the `redux-thunk` package. Even the
   * middleware will eventually dispatch plain object actions using this method.
   *
   * @param {Object} action A plain object representing “what changed”. It is
   * a good idea to keep actions serializable so you can record and replay user
   * sessions, or use the time travelling `redux-devtools`. An action must have
   * a `type` property which may not be `undefined`. It is a good idea to use
   * string constants for action types.
   *
   * @returns {Object} For convenience, the same action object you dispatched.
   *
   * Note that, if you use a custom middleware, it may wrap `dispatch()` to
   * return something else (for example, a Promise you can await).
   */


  function dispatch(action) {
    if (!isPlainObject(action)) {
      throw new Error('Actions must be plain objects. ' + 'Use custom middleware for async actions.');
    }

    if (typeof action.type === 'undefined') {
      throw new Error('Actions may not have an undefined "type" property. ' + 'Have you misspelled a constant?');
    }

    if (isDispatching) {
      throw new Error('Reducers may not dispatch actions.');
    }

    try {
      isDispatching = true;
      currentState = currentReducer(currentState, action);
    } finally {
      isDispatching = false;
    }

    var listeners = currentListeners = nextListeners;

    for (var i = 0; i < listeners.length; i++) {
      var listener = listeners[i];
      listener();
    }

    return action;
  }
  /**
   * Replaces the reducer currently used by the store to calculate the state.
   *
   * You might need this if your app implements code splitting and you want to
   * load some of the reducers dynamically. You might also need this if you
   * implement a hot reloading mechanism for Redux.
   *
   * @param {Function} nextReducer The reducer for the store to use instead.
   * @returns {void}
   */


  function replaceReducer(nextReducer) {
    if (typeof nextReducer !== 'function') {
      throw new Error('Expected the nextReducer to be a function.');
    }

    currentReducer = nextReducer;
    dispatch({
      type: ActionTypes.REPLACE
    });
  }
  /**
   * Interoperability point for observable/reactive libraries.
   * @returns {observable} A minimal observable of state changes.
   * For more information, see the observable proposal:
   * https://github.com/tc39/proposal-observable
   */


  function observable() {
    var _ref;

    var outerSubscribe = subscribe;
    return _ref = {
      /**
       * The minimal observable subscription method.
       * @param {Object} observer Any object that can be used as an observer.
       * The observer object should have a `next` method.
       * @returns {subscription} An object with an `unsubscribe` method that can
       * be used to unsubscribe the observable from the store, and prevent further
       * emission of values from the observable.
       */
      subscribe: function subscribe(observer) {
        if (typeof observer !== 'object' || observer === null) {
          throw new TypeError('Expected the observer to be an object.');
        }

        function observeState() {
          if (observer.next) {
            observer.next(getState());
          }
        }

        observeState();
        var unsubscribe = outerSubscribe(observeState);
        return {
          unsubscribe: unsubscribe
        };
      }
    }, _ref[result] = function () {
      return this;
    }, _ref;
  } // When a store is created, an "INIT" action is dispatched so that every
  // reducer returns their initial state. This effectively populates
  // the initial state tree.


  dispatch({
    type: ActionTypes.INIT
  });
  return _ref2 = {
    dispatch: dispatch,
    subscribe: subscribe,
    getState: getState,
    replaceReducer: replaceReducer
  }, _ref2[result] = observable, _ref2;
}

/**
 * Prints a warning in the console if it exists.
 *
 * @param {String} message The warning message.
 * @returns {void}
 */
function warning(message) {
  /* eslint-disable no-console */
  if (typeof console !== 'undefined' && typeof console.error === 'function') {
    console.error(message);
  }
  /* eslint-enable no-console */


  try {
    // This error was thrown as a convenience so that if you enable
    // "break on all exceptions" in your console,
    // it would pause the execution at this line.
    throw new Error(message);
  } catch (e) {} // eslint-disable-line no-empty

}

function getUndefinedStateErrorMessage(key, action) {
  var actionType = action && action.type;
  var actionDescription = actionType && "action \"" + String(actionType) + "\"" || 'an action';
  return "Given " + actionDescription + ", reducer \"" + key + "\" returned undefined. " + "To ignore an action, you must explicitly return the previous state. " + "If you want this reducer to hold no value, you can return null instead of undefined.";
}

function getUnexpectedStateShapeWarningMessage(inputState, reducers, action, unexpectedKeyCache) {
  var reducerKeys = Object.keys(reducers);
  var argumentName = action && action.type === ActionTypes.INIT ? 'preloadedState argument passed to createStore' : 'previous state received by the reducer';

  if (reducerKeys.length === 0) {
    return 'Store does not have a valid reducer. Make sure the argument passed ' + 'to combineReducers is an object whose values are reducers.';
  }

  if (!isPlainObject(inputState)) {
    return "The " + argumentName + " has unexpected type of \"" + {}.toString.call(inputState).match(/\s([a-z|A-Z]+)/)[1] + "\". Expected argument to be an object with the following " + ("keys: \"" + reducerKeys.join('", "') + "\"");
  }

  var unexpectedKeys = Object.keys(inputState).filter(function (key) {
    return !reducers.hasOwnProperty(key) && !unexpectedKeyCache[key];
  });
  unexpectedKeys.forEach(function (key) {
    unexpectedKeyCache[key] = true;
  });
  if (action && action.type === ActionTypes.REPLACE) return;

  if (unexpectedKeys.length > 0) {
    return "Unexpected " + (unexpectedKeys.length > 1 ? 'keys' : 'key') + " " + ("\"" + unexpectedKeys.join('", "') + "\" found in " + argumentName + ". ") + "Expected to find one of the known reducer keys instead: " + ("\"" + reducerKeys.join('", "') + "\". Unexpected keys will be ignored.");
  }
}

function assertReducerShape(reducers) {
  Object.keys(reducers).forEach(function (key) {
    var reducer = reducers[key];
    var initialState = reducer(undefined, {
      type: ActionTypes.INIT
    });

    if (typeof initialState === 'undefined') {
      throw new Error("Reducer \"" + key + "\" returned undefined during initialization. " + "If the state passed to the reducer is undefined, you must " + "explicitly return the initial state. The initial state may " + "not be undefined. If you don't want to set a value for this reducer, " + "you can use null instead of undefined.");
    }

    if (typeof reducer(undefined, {
      type: ActionTypes.PROBE_UNKNOWN_ACTION()
    }) === 'undefined') {
      throw new Error("Reducer \"" + key + "\" returned undefined when probed with a random type. " + ("Don't try to handle " + ActionTypes.INIT + " or other actions in \"redux/*\" ") + "namespace. They are considered private. Instead, you must return the " + "current state for any unknown actions, unless it is undefined, " + "in which case you must return the initial state, regardless of the " + "action type. The initial state may not be undefined, but can be null.");
    }
  });
}
/**
 * Turns an object whose values are different reducer functions, into a single
 * reducer function. It will call every child reducer, and gather their results
 * into a single state object, whose keys correspond to the keys of the passed
 * reducer functions.
 *
 * @param {Object} reducers An object whose values correspond to different
 * reducer functions that need to be combined into one. One handy way to obtain
 * it is to use ES6 `import * as reducers` syntax. The reducers may never return
 * undefined for any action. Instead, they should return their initial state
 * if the state passed to them was undefined, and the current state for any
 * unrecognized action.
 *
 * @returns {Function} A reducer function that invokes every reducer inside the
 * passed object, and builds a state object with the same shape.
 */


function combineReducers(reducers) {
  var reducerKeys = Object.keys(reducers);
  var finalReducers = {};

  for (var i = 0; i < reducerKeys.length; i++) {
    var key = reducerKeys[i];

    if (process.env.NODE_ENV !== 'production') {
      if (typeof reducers[key] === 'undefined') {
        warning("No reducer provided for key \"" + key + "\"");
      }
    }

    if (typeof reducers[key] === 'function') {
      finalReducers[key] = reducers[key];
    }
  }

  var finalReducerKeys = Object.keys(finalReducers);
  var unexpectedKeyCache;

  if (process.env.NODE_ENV !== 'production') {
    unexpectedKeyCache = {};
  }

  var shapeAssertionError;

  try {
    assertReducerShape(finalReducers);
  } catch (e) {
    shapeAssertionError = e;
  }

  return function combination(state, action) {
    if (state === void 0) {
      state = {};
    }

    if (shapeAssertionError) {
      throw shapeAssertionError;
    }

    if (process.env.NODE_ENV !== 'production') {
      var warningMessage = getUnexpectedStateShapeWarningMessage(state, finalReducers, action, unexpectedKeyCache);

      if (warningMessage) {
        warning(warningMessage);
      }
    }

    var hasChanged = false;
    var nextState = {};

    for (var _i = 0; _i < finalReducerKeys.length; _i++) {
      var _key = finalReducerKeys[_i];
      var reducer = finalReducers[_key];
      var previousStateForKey = state[_key];
      var nextStateForKey = reducer(previousStateForKey, action);

      if (typeof nextStateForKey === 'undefined') {
        var errorMessage = getUndefinedStateErrorMessage(_key, action);
        throw new Error(errorMessage);
      }

      nextState[_key] = nextStateForKey;
      hasChanged = hasChanged || nextStateForKey !== previousStateForKey;
    }

    return hasChanged ? nextState : state;
  };
}

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function _objectSpread(target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i] != null ? arguments[i] : {};
    var ownKeys = Object.keys(source);

    if (typeof Object.getOwnPropertySymbols === 'function') {
      ownKeys = ownKeys.concat(Object.getOwnPropertySymbols(source).filter(function (sym) {
        return Object.getOwnPropertyDescriptor(source, sym).enumerable;
      }));
    }

    ownKeys.forEach(function (key) {
      _defineProperty(target, key, source[key]);
    });
  }

  return target;
}

/**
 * Composes single-argument functions from right to left. The rightmost
 * function can take multiple arguments as it provides the signature for
 * the resulting composite function.
 *
 * @param {...Function} funcs The functions to compose.
 * @returns {Function} A function obtained by composing the argument functions
 * from right to left. For example, compose(f, g, h) is identical to doing
 * (...args) => f(g(h(...args))).
 */
function compose() {
  for (var _len = arguments.length, funcs = new Array(_len), _key = 0; _key < _len; _key++) {
    funcs[_key] = arguments[_key];
  }

  if (funcs.length === 0) {
    return function (arg) {
      return arg;
    };
  }

  if (funcs.length === 1) {
    return funcs[0];
  }

  return funcs.reduce(function (a, b) {
    return function () {
      return a(b.apply(void 0, arguments));
    };
  });
}

/**
 * Creates a store enhancer that applies middleware to the dispatch method
 * of the Redux store. This is handy for a variety of tasks, such as expressing
 * asynchronous actions in a concise manner, or logging every action payload.
 *
 * See `redux-thunk` package as an example of the Redux middleware.
 *
 * Because middleware is potentially asynchronous, this should be the first
 * store enhancer in the composition chain.
 *
 * Note that each middleware will be given the `dispatch` and `getState` functions
 * as named arguments.
 *
 * @param {...Function} middlewares The middleware chain to be applied.
 * @returns {Function} A store enhancer applying the middleware.
 */

function applyMiddleware() {
  for (var _len = arguments.length, middlewares = new Array(_len), _key = 0; _key < _len; _key++) {
    middlewares[_key] = arguments[_key];
  }

  return function (createStore) {
    return function () {
      var store = createStore.apply(void 0, arguments);

      var _dispatch = function dispatch() {
        throw new Error("Dispatching while constructing your middleware is not allowed. " + "Other middleware would not be applied to this dispatch.");
      };

      var middlewareAPI = {
        getState: store.getState,
        dispatch: function dispatch() {
          return _dispatch.apply(void 0, arguments);
        }
      };
      var chain = middlewares.map(function (middleware) {
        return middleware(middlewareAPI);
      });
      _dispatch = compose.apply(void 0, chain)(store.dispatch);
      return _objectSpread({}, store, {
        dispatch: _dispatch
      });
    };
  };
}

/*
 * This is a dummy function to check if the function name has been altered by minification.
 * If the function has been minified and NODE_ENV !== 'production', warn the user.
 */

function isCrushed() {}

if (process.env.NODE_ENV !== 'production' && typeof isCrushed.name === 'string' && isCrushed.name !== 'isCrushed') {
  warning('You are currently using minified code outside of NODE_ENV === "production". ' + 'This means that you are running a slower development build of Redux. ' + 'You can use loose-envify (https://github.com/zertosh/loose-envify) for browserify ' + 'or setting mode to production in webpack (https://webpack.js.org/concepts/mode/) ' + 'to ensure you have the correct code for your production build.');
}

var ReactReduxOnTheFly = function () {
  function ReactReduxOnTheFly(staticReducers) {
    var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    classCallCheck(this, ReactReduxOnTheFly);

    this._staticReducers = staticReducers;
    this._options = options;
    this.MAX_ASYNC_REDUCERS_STACK = this._options.maxReducersNumber || 30;
  }

  createClass(ReactReduxOnTheFly, [{
    key: "asyncCombineReducers",
    value: function asyncCombineReducers(asyncReducers) {
      return combineReducers(_extends({}, this._staticReducers, asyncReducers));
    }
  }, {
    key: "asyncReducerManager",
    value: function asyncReducerManager(store) {
      var asyncReducersStack = Object.keys(store.asyncReducers);

      if (asyncReducersStack.length > this.MAX_ASYNC_REDUCERS_STACK) {
        this.ejectAsyncReducer(store, asyncReducersStack[0]);
      }
    }

    /**
     * injectAsyncReducer
     * @param {object} store
     * @param {string} name
     * @param {function} asyncReducer
     */

  }, {
    key: "injectAsyncReducer",
    value: function injectAsyncReducer(store, name, asyncReducer) {
      store.asyncReducers[name] = asyncReducer;
      store.replaceReducer(this.asyncCombineReducers(store.asyncReducers));
    }
  }, {
    key: "ejectAsyncReducer",
    value: function ejectAsyncReducer(store, name) {
      delete store.asyncReducers[name];
      store.replaceReducer(this.asyncCombineReducers(store.asyncReducers));
    }
  }]);
  return ReactReduxOnTheFly;
}();

/* eslint-disable react/prop-types */

/**
 * createRoot
 * @param Provider - React-redux provider
 * @param {object} [staticReducers]
 * @param {array} [middleware]
 * @param {*} [initialState]
 */
var createRoot = function createRoot(Provider, staticReducers, middleware, initialState) {
  var reactReduxOnTheFly = new ReactReduxOnTheFly(staticReducers);
  var reducers = reactReduxOnTheFly.asyncCombineReducers();
  var store = createStore(reducers, initialState, applyMiddleware.apply(undefined, toConsumableArray(middleware || [])));
  store.asyncReducers = {};

  return function (props) {
    return React.createElement(
      Provider,
      { store: store },
      React.createElement(
        ReactReduxOnTheFlyProvider,
        {
          store: store,
          reactReduxOnTheFly: reactReduxOnTheFly
        },
        props.children
      )
    );
  };
};

export { ReactReduxOnTheFlyProvider, ReactReduxOnTheFly, withAsyncReducer, createActions, createNamedWrapperReducer, baseArrayReducer, baseObjectReducer, createDuck, createRoot };
//# sourceMappingURL=index.es.js.map
